package cerebral.parallelCoordinates.cluster;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import prefuse.Visualization;
import prefuse.visual.VisualItem;
import cerebral.parallelCoordinates.ParallelCoordinates;

public class KMeansCluster {

	private int targetK;

	private int k;

	private Visualization vis;

	private int dataDimensions;
    private Cluster[] clusters;
    
	public KMeansCluster(Visualization vis, int dataDimensions) {
		this.vis = vis;
		this.dataDimensions = dataDimensions;
	}

	public void cluster(int targetK) {

		this.targetK = targetK;

		List items = new ArrayList();
		Iterator dataItems = vis.items(ParallelCoordinates.DATA);
		while (dataItems.hasNext()) {
			VisualItem dataItem = (VisualItem) dataItems.next();
			if (!dataItem.getBoolean(ParallelCoordinates.IGNORE) && dataItem.isVisible()) {
				items.add(dataItem);
			}
		}

		if (targetK > items.size()) {
			k = items.size();
		} else {
			k = targetK;
		}
		if (k < 1) {
			k = 0;
			return;
		}
        clusters = new Cluster[k];
        for (int i=0; i < k; i++){
            clusters[i] = new Cluster();
        }
        
		int[] seeds = new int[k];
		Random r = new Random();
		// choose centroids at random
		for (int i = 0; i < k; i++) {
			boolean uniqueR = true;
			int seed;
			do {
				uniqueR = true;
				seed = r.nextInt(items.size());
				for (int j = 0; j < i; j++) {
					if (seeds[j] == seed) {
						uniqueR = false;
						break;
					}
				}
				seeds[i] = seed;
			} while (!uniqueR);

			VisualItem seedItem = (VisualItem) items.get(seed);
			List values = (List) seedItem.get(ParallelCoordinates.NORM_VALUES);
			Iterator valuesIt = values.iterator();
			while (valuesIt.hasNext()) {
				clusters[i].center.add(valuesIt.next());
			}
		}

		double maxMove = 0.0;
		do {
			maxMove = 0.0;
			// find closest centroid for each data point
			for (int i = 0; i < k; i++) {
				clusters[i].members.clear();
			}

			Iterator itemsIt = items.iterator();
			while (itemsIt.hasNext()) {
				VisualItem item = (VisualItem) itemsIt.next();
				int minDistIndex = 0;
				double minDistance = 1000000.0;
				for (int j = 0; j < k; j++) {
					double dist = distance((List) item
							.get(ParallelCoordinates.NORM_VALUES), clusters[j].center);
					if (dist < minDistance) {
						minDistance = dist;
						minDistIndex = j;
					}
				}
				clusters[minDistIndex].members.add(item);
			}

			// recalculate centroids
			for (int i = 0; i < k; i++) {
				double[] newCentroid = new double[dataDimensions];
				Arrays.fill(newCentroid, 0.0);
				Iterator it = clusters[i].members.iterator();
				while (it.hasNext()) {
					VisualItem member = (VisualItem) it.next();
					List values = (List) member.get(ParallelCoordinates.NORM_VALUES);
					for (int j = 0; j < values.size(); j++) {
						newCentroid[j] += ((Double) values.get(j))
								.doubleValue();
					}
				}
				
				for (int j = 0; j < dataDimensions; j++) {
					newCentroid[j] /= clusters[i].members.size();
				}

				double movement = distance(newCentroid, clusters[i].center);
				if (movement > maxMove) {
					maxMove = movement;
				}

				clusters[i].center.clear();
				for (int j = 0; j < dataDimensions; j++) {
					clusters[i].center.add(new Double(newCentroid[j]));
				}
			}
		} while (maxMove > 0.0005);

		Arrays.sort(clusters, new Comparator() {

			public int compare(Object o1, Object o2) {
				Cluster a = (Cluster) o1;
				Cluster b = (Cluster) o2;
				if (a.members.size() < b.members.size()) {
					return 1;
				} else if (a.members.size() == b.members.size()) {
					return 0;
				} else {
					return -1;
				}
			}

		});
	}

	private double distance(double[] a, List b) {
	   

		double d = 0.0;
		for (int i = 0; i < b.size(); i++) {
			double aVal = a[i];
			double bVal = ((Double) b.get(i)).doubleValue();
			d += (aVal - bVal) * (aVal - bVal);
		}
		return d;
		
	    /*
 //Let's calculate the pearson coefficient instead
 //no wait.  Let's not.  Let's normalize the data first
  //and then use standard euclidean distance.
  
        
        double sumXY = 0.0;
        double sumX = 0.0;
        double sumXX = 0.0;
        double sumY = 0.0;
        double sumYY = 0.0;
        double n = a.length;
        for (int i = 0; i < a.length; i++) {
            double aVal = a[i];
            double bVal = ((Double) b.get(i)).doubleValue();
            sumXY += (aVal * bVal);
            sumX += aVal;
            sumXX += aVal * aVal;
            sumY += bVal;
            sumYY += bVal*bVal;            
        }
        double numerator = sumXY -  ((sumX * sumY)/n);
        double denomX = sumXX - ((sumXX)/n);
        double denomY = sumYY - ((sumYY)/n);
        double denominator = Math.sqrt(denomX*denomY);
        return numerator/denominator;
        */
	}

	private double distance(List a, List b) {
	    
	    
		double d = 0.0;
		for (int i = 0; i < a.size(); i++) {
			double aVal = ((Double) a.get(i)).doubleValue();
			double bVal = ((Double) b.get(i)).doubleValue();
			d += (aVal - bVal) * (aVal - bVal);
		}
		return d;
		        
	}

	public List getClusterCenter(int i) {
		return clusters[i].center;
	}

	public List getMembers(int i) {
		return clusters[i].members;
	}

	public int getK() {
		return k;
	}

	public void recluster() {
		cluster(targetK);
	}
    
    private class Cluster {
        public List center;
        public List members;
        public Cluster(){
            center = new ArrayList();
            members = new ArrayList();
        }
    }

}
